export default class Schema {
  constructor(db) {
    this.db = db;
  }
  execSQL(sql) {
    console.debug(sql);
    return this.db.executeSql(sql);
  }

  /**
   * Crea la tabla
   * @param tableName 
   * @param closure 
   * @returns 
   */
  async create(tableName, closure) {
    const table = new Table(tableName);
    await closure(table);

    // Create
    await this.execSQL(table.toString());
    table.isCreated = true;
    for (const index of table.indexs) {
      await this.execSQL(index.toString());
    }
    return table;
  }

  /**
   * Agrega las columnas nuevas que no existen fisicamente en la tabla.
   * @param tableName 
   * @param closure 
   */
  async alter(tableName, closure) {
    const table = new Table(tableName);
    await closure(table);

    // Alter
    const set = await this.getColumnsNotExist(table);
    for (let i = 0; i < set.length; i++) {
      const col = set[i];
      await this.execSQL(`
              ALTER TABLE ${table.name} ADD COLUMN ${col}
          `);
    }
    for (const index of table.indexs) {
      await this.execSQL(`DROP INDEX IF EXISTS ${index.name}`);
      await this.execSQL(index.toString());
    }
    return table;
  }

  /**
   * Crea o modifica la tabla
   * @param tableName 
   * @param closure 
   * @returns 
   */
  async createOrAlter(tableName, closure) {
    if (await this.hasTable(tableName)) {
      return await this.alter(tableName, closure);
    } else {
      return await this.create(tableName, closure);
    }
  }

  /**
   * Valida si existe una tabla en la base de datos.
   */
  async hasTable(tableName) {
    const {
      rows
    } = await this.execSQL(`
            SELECT COUNT(*) AS _count_ FROM sqlite_master WHERE type = 'table' AND name = '${tableName}'
        `);
    if (rows[0]) {
      const first = rows[0];
      return first['_count_'] > 0;
    }
    return false;
  }

  /**
   * Obtiene un lista de columnas que ha un no exiten fisicamente en la tabla.
   */
  async getColumnsNotExist(table) {
    const {
      rows
    } = await this.execSQL(`
            PRAGMA table_info(${table.name})
        `);
    const existingColumnNames = rows.map(it => it["name"]);
    const missingColumns = [];
    for (const col of table.columns) {
      // Verifica si la columna no está presente
      if (!existingColumnNames.includes(col.name)) {
        // La columna no existe, añadirla al resultado
        missingColumns.push(col);
      }
    }
    return missingColumns;
  }
}

/*
 * CREATE TABLE IF NOT EXISTS tb_usuario ( ... )
 */
export class Table {
  isCreated = false;
  columns = new Array();
  constraints = new Array();
  indexs = new Array();
  constructor(name) {
    this.name = name;
  }
  addColumn(columname, type) {
    const col = new Column(columname, type);
    this.columns.push(col);
    return col;
  }
  addConstraint() {
    const constraint = new Constraint();
    this.constraints.push(constraint);
    return constraint;
  }
  addIndex(indexname) {
    const index = new Index(indexname, this.name);
    this.indexs.push(index);
    return index;
  }

  /** Incrementing ID to the table (primary key)  */
  increments(column) {
    return this.integer(column).primaryKey().autoIncrement();
  }

  /** INTEGER equivalent to the table  */
  integer(column) {
    return this.addColumn(column, "INTEGER");
  }

  /** FLOAT equivalent to the table  */
  float(column) {
    return this.addColumn(column, "FLOAT");
  }

  /** DECIMAL equivalent with a precision and scale  */
  real(column) {
    return this.addColumn(column, "REAL");
  }

  /** VARCHAR equivalent column  */
  text(column) {
    return this.addColumn(column, "TEXT");
  }

  /** Binary equivalent column  */
  blob(column) {
    return this.addColumn(column, "BLOB");
  }
  toString() {
    // Asegura que las columnas y las restricciones se unan correctamente, separadas por comas si es necesario
    const columnDefinitions = [...this.columns, ...this.constraints].join(",\n\t");
    return `
        CREATE TABLE IF NOT EXISTS ${this.name} (
          ${columnDefinitions})
      `;
  }
  foreign(columnname) {
    return this.addConstraint().foreign(columnname);
  }
  index(indexname) {
    return this.addIndex(indexname);
  }
}

/**
 * id INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL DEFAUL '' +
 */
export class Column {
  _primaryKey = false;
  _autoIncrement = false;
  _nullable = false;
  _defaultVal = null;
  constructor(name, type) {
    this.name = name;
    this.type = type;
  }
  primaryKey() {
    this._primaryKey = true;
    return this;
  }
  autoIncrement() {
    this._autoIncrement = true;
    return this;
  }
  default(defaultVal) {
    this._defaultVal = defaultVal;
    return this;
  }
  nullable() {
    this._nullable = true;
    return this;
  }
  toString() {
    let sql = `${this.name} ${this.type}`;
    if (this._primaryKey) sql += " PRIMARY KEY";
    if (this._autoIncrement) sql += " AUTOINCREMENT";
    sql += this._nullable ? " NULL" : " NOT NULL";
    if (this._defaultVal !== null) {
      // Aquí se simplifica el escape de la cadena SQL. 
      // Deberías implementar una función adecuada para escapar la cadena SQL según tus necesidades.
      sql += ` DEFAULT '${this._defaultVal}'`;
    }
    return sql;
  }
}

/**
 * CREATE UNIQUE INDEX IF NOT EXISTS 'index_picking'
 * ON 'tb_picking' ('folio')
 */
export class Index {
  _unique = false;
  _columns = [];
  constructor(name, tablename) {
    this.name = name;
    this.tablename = tablename;
  }
  unique(...columns) {
    this._unique = true;
    this._columns = columns;
  }
  columns(...columns) {
    this._columns = columns;
  }
  toString() {
    let sql = `CREATE`;
    if (this._unique) sql += ` UNIQUE`;
    sql += ` INDEX IF NOT EXISTS ${this.escapeSQLString(this.name)} ON ${this.escapeSQLString(this.tablename)} (`;
    sql += this._columns.map(column => `${this.escapeSQLString(column)}`).join(", ");
    sql += `)`;
    return sql;
  }
  escapeSQLString(value) {
    // Implementa aquí tu lógica de escape de cadenas SQL, según sea necesario.
    // Esta es una implementación básica que simplemente devuelve la cadena entre comillas simples como ejemplo.
    return `'${value.replace(/'/g, "''")}'`;
  }
}

/**
 * FOREIGN KEY(id_inventario) REFERENCES tb_inventario(id) ON DELETE CASCADE
 */
export class Constraint {
  _foreign = [];
  _references = [];
  _table = null;
  _onDelete = null;
  foreign(...key) {
    this._foreign = key;
    return this;
  }
  references(...key) {
    this._references = key;
    return this;
  }
  on(table) {
    this._table = table;
    return this;
  }
  onDelete(on = "CASCADE") {
    this._onDelete = on;
  }
  toString() {
    let sql = "";
    if (this._foreign.length > 0) sql += `FOREIGN KEY (${this._foreign.join(", ")})`;
    if (this._table !== null && this._references.length > 0) sql += ` REFERENCES ${this._table} (${this._references.join(", ")})`;
    if (this._onDelete !== null) sql += ` ON DELETE ${this._onDelete}`;
    return sql;
  }
}
//# sourceMappingURL=Schema.js.map